import org.junit.Test;

import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class BasicTests {

    private static final int APROX = 100;
    private static double ASSERT_DELTA = 0.1;
    private static double[] solutions = new double[]{70.04, 46.28, 42.18, 51.56, 70.97};
    private static int[] solutionsIter = new int[]{344, 320, 167, 447, 489};
    private static double[] solutionsSlow = new double[]{38.28, 111.88, 77.74, 85.5};

    private static boolean checkCoordinateValid(int cord, int max) {
        return cord >= 0 && cord < max;
    }

    private static int countNumOfState(FactoryMap factoryMap, FactoryMap.MapPoint.State state) {
        int count = 0;
        for (int xCor = 0; xCor < factoryMap.getSizeX(); xCor++) {
            for (int yCor = 0; yCor < factoryMap.getSizeY(); yCor++) {
                if (factoryMap.getPoint(xCor, yCor).getState() == state)
                    count++;
            }
        }
        return count;
    }

    private static double computeSingleHopDistance(FactoryMap.MapPoint point1, FactoryMap.MapPoint point2, int sizeX, int sizeY) {
        return computeSingleHopDistance(point1, point2, sizeX, sizeY, false);
    }

    private static double computeSingleHopDistance(FactoryMap.MapPoint point1, FactoryMap.MapPoint point2, int sizeX, int sizeY, boolean slowed) {
        double slowingFactor = 2.0;
        assertTrue("Coordinates in path are not valid",
                checkCoordinateValid(point2.x, sizeX) &&
                        checkCoordinateValid(point1.x, sizeX) &&
                        checkCoordinateValid(point2.y, sizeY) &&
                        checkCoordinateValid(point1.y, sizeY));
        double xD = point2.x - point1.x;
        double yD = point2.y - point1.y;
        assertTrue("Shortest path is not valid", (int) Math.abs(xD) <= 1);
        assertTrue("Shortest path is not valid", (int) Math.abs(yD) <= 1);
        double dist = computeGeomDist(point1, point2);
        if (slowed) {
            dist += (point1.isSlowed()) ? slowingFactor : 0.0;
            dist += (point2.isSlowed()) ? slowingFactor : 0.0;
        }
        return dist;
    }

    private static double computePathDistance(List<FactoryMap.MapPoint> path, int sizeX, int sizeY, boolean slowed) {
        double dist = 0;
        for (int i = 0; i < path.size() - 1; i++) {
            dist += computeSingleHopDistance(path.get(i), path.get(i + 1), sizeX, sizeY, slowed);
        }
        return dist;
    }

    private static double computePathDistance(List<FactoryMap.MapPoint> path, int sizeX, int sizeY) {
        return computePathDistance(path, sizeX, sizeY, false);
    }

    private static double computeGeomDist(FactoryMap.MapPoint p1, FactoryMap.MapPoint p2) {
        double xD = p2.x - p1.x;
        double yD = p2.y - p1.y;
        return Math.sqrt(xD * xD + yD * yD);
    }

    private static RoutePlanner testShortestPath(int seed, int sizeX, int sizeY, int numObstacles, double solution, boolean doApprox) throws RoutePlannerException {
        FactoryMap factoryMap = FactoryMapUtil.createRandomMap(sizeX, sizeY, numObstacles, seed);
        return testShortestPath(factoryMap, solution, doApprox, false);
    }

    private static RoutePlanner testShortestPath(FactoryMap factoryMap, double solution, boolean doApprox, boolean slowed) throws RoutePlannerException {
        RoutePlanner task = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0), factoryMap.getPoint(factoryMap.getSizeX() - 1, factoryMap.getSizeY() - 1), doApprox);
        RouteUtil.computeShortestPath(task, null, 1);
        System.out.println("\n" + factoryMap);
        List<FactoryMap.MapPoint> path = task.getShortestPath();
        assertEquals("The path is either illegal or not the shortest path", solution, computePathDistance(path, factoryMap.getSizeX(), factoryMap.getSizeY(), slowed), ASSERT_DELTA);
        return task;
    }

    private static RoutePlanner testShortestPathSteps(int seed, int sizeX, int sizeY, int numObstacles, int solID, boolean doApprox) throws RoutePlannerException {
        FactoryMap factoryMap = FactoryMapUtil.createRandomMap(sizeX, sizeY, numObstacles, seed);
        RoutePlanner task = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0), factoryMap.getPoint(sizeX - 1, sizeY - 1), doApprox);
        int count = RouteUtil.computeShortestPath(task, null, 1);
        assertTrue("Too many steps to find the shortest path " + count, solutionsIter[solID] - APROX <= count && count <= solutionsIter[solID] + APROX);
        return task;
    }

    @Test
    public void checkMapMarked() throws RoutePlannerException {
        int sizeX = 25;
        int sizeY = 25;
        FactoryMap factoryMap = FactoryMapUtil.createDemoMap();
        RoutePlanner task = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0), factoryMap.getPoint(sizeX - 1, sizeY - 1));

        int count = 0;
        while (!task.doNextStep()) {
            int numExplored = countNumOfState(factoryMap, FactoryMap.MapPoint.State.EXPLORED);
            assertEquals("No node was explored on the map on a nextStep invocation", ++count, numExplored);
        }
    }

    @Test
    public void checkNeighbours() throws RoutePlannerException {
        FactoryMap factoryMap = new FactoryMap(3, 3);
        List<FactoryMap.MapPoint> sol = new ArrayList<>();
        for (int xCor = 0; xCor < 3; xCor++) {
            for (int yCor = 0; yCor < 3; yCor++) {
                if (!(xCor == 1 && yCor == 1)) {
                    sol.add(factoryMap.getPoint(xCor, yCor));
                }
            }
        }
        for (int xCor = 0; xCor < 3; xCor++) {
            for (int yCor = 0; yCor < 3; yCor++) {
                if (!(xCor == 1 && yCor == 1)) {
                    factoryMap.getPoint(xCor, yCor).setBlocked();
                    sol.remove(factoryMap.getPoint(xCor, yCor));
                    List<FactoryMap.MapPoint> points = factoryMap.getMovableNeighbours(factoryMap.getPoint(1, 1));
                    assertEquals("A node is not an neighbour of itself", sol.size(), points.size());
                    for (FactoryMap.MapPoint solPoint : sol)
                        assertTrue("getMovableNeighbours does not return all valid neighbours",
                                points.contains(solPoint));
                }

            }
        }
        // check bounds
        factoryMap = new FactoryMap(3, 3);
        List<FactoryMap.MapPoint> points = factoryMap.getMovableNeighbours(factoryMap.getPoint(2, 2));
        assertEquals("Invalid nodes are returned in getMovableNeighbours", 3, points.size());
        assertTrue("getMovableNeighbours does not return all valid neighbours",
                points.contains(factoryMap.getPoint(1, 2)));
        assertTrue("getMovableNeighbours does not return all valid neighbours",
                points.contains(factoryMap.getPoint(1, 1)));
        assertTrue("getMovableNeighbours does not return all valid neighbours",
                points.contains(factoryMap.getPoint(2, 1)));
    }

    @Test
    public void distanceTest() throws RoutePlannerException {
        FactoryMap factoryMap = new FactoryMap(10, 10);
        FactoryMap.MapPoint from = factoryMap.getPoint(4, 4);
        for (int xCor = 0; xCor < 10; xCor++) {
            for (int yCor = 0; yCor < 10; yCor++) {
                FactoryMap.MapPoint to = factoryMap.getPoint(xCor, yCor);
                assertEquals("MapPoint.getDistance() is not correct", computeGeomDist(from, to), from.getDistance(to), ASSERT_DELTA);

            }
        }
    }

    @Test
    public void checkSlowedDistance() throws RoutePlannerException {
        double slowingFactor = 2.0;
        FactoryMap factoryMap = FactoryMapUtil.createDemoMapSlow();
        FactoryMap.MapPoint from = factoryMap.getPoint(4, 4);
        for (int xCor = 0; xCor < 10; xCor++) {
            for (int yCor = 0; yCor < 10; yCor++) {
                FactoryMap.MapPoint to = factoryMap.getPoint(xCor, yCor);
                double dist = computeGeomDist(from, to);
                dist += (from.isSlowed()) ? slowingFactor : 0.0;
                dist += (to.isSlowed()) ? slowingFactor : 0.0;
                assertEquals("MapPoint.getDistanceWithSlowingFactor() is not correct", dist, from.getDistanceWithSlowingFactor(to, slowingFactor), ASSERT_DELTA);

            }
        }
    }

    @Test
    public void checkDemoMap() throws RoutePlannerException {
        FactoryMap factoryMap = FactoryMapUtil.createDemoMap();
        RoutePlanner task = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0), factoryMap.getPoint(factoryMap.getSizeX() - 1, factoryMap.getSizeY() - 1));
        RouteUtil.computeShortestPath(task, null, 1);
        List<FactoryMap.MapPoint> path = task.getShortestPath();
        assertEquals("The path is either illegal or not the shortest path", solutions[0], computePathDistance(path, 25, 25), ASSERT_DELTA);

    }

    @Test
    public void checkAlgorithmStates() throws RoutePlannerException {
        FactoryMap factoryMap = FactoryMapUtil.createDemoMap();
        RoutePlanner task = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0), factoryMap.getPoint(factoryMap.getSizeX() - 1, factoryMap.getSizeY() - 1));
        GraphNode current = task.getCurrentPosition();
        GraphNode next = task.getCurrentPosition();
        while (!task.doNextStep()) {
            next = task.getCurrentPosition();
            assertTrue("Algorithm selects same node ", !next.equals(current));
            current = next;
        }
    }

    @Test
    public void checkSlowDemoMap() throws RoutePlannerException {
        FactoryMap factoryMap = FactoryMapUtil.createDemoMapSlow();
        RoutePlanner task = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0), factoryMap.getPoint(factoryMap.getSizeX() - 1, factoryMap.getSizeY() - 1));
        RouteUtil.computeShortestPath(task, null, 1);
        List<FactoryMap.MapPoint> path = task.getShortestPath();
        assertEquals("The path is either illegal or not the shortest path", solutionsSlow[0], computePathDistance(path, 25, 25, true), ASSERT_DELTA);

    }


    @Test
    public void testShortestPath1Rand() throws RoutePlannerException {
        testShortestPath(1111, 30, 30, 4, solutions[1], true);
    }

    @Test
    public void testShortestPath2Rand() throws RoutePlannerException {
        testShortestPath(2222, 30, 30, 4, solutions[2], true);
    }

    @Test
    public void testShortestPath3Rand() throws RoutePlannerException {
        testShortestPath(3333, 30, 30, 4, solutions[3], true);
    }

    @Test
    public void testShortestPath4Rand() throws RoutePlannerException {
        testShortestPath(4444, 40, 40, 8, solutions[4], true);
    }

    @Test
    public void testShortestPathWithApproximation1() throws RoutePlannerException {
        testShortestPathSteps(3333, 30, 30, 4, 3, true);
    }

    @Test
    public void testShortestPathWithApproximation2() throws RoutePlannerException {
        testShortestPathSteps(4444, 40, 40, 8, 4, true);
    }


    @Test
    public void testShortestPath1RandSlowed() throws RoutePlannerException {
        testShortestPath(FactoryMapUtil.createRandomMap(50, 50, 6, 111, 1200), solutionsSlow[1], true, true);
    }

    @Test
    public void testShortestPath2RandSlowed() throws RoutePlannerException {
        testShortestPath(FactoryMapUtil.createRandomMap(50, 50, 6, 222, 1200), solutionsSlow[2], true, true);
    }

    @Test
    public void testShortestPath3RandSlowed() throws RoutePlannerException {
        testShortestPath(FactoryMapUtil.createRandomMap(50, 50, 6, 333, 1200), solutionsSlow[3], true, true);
    }


    @Test
    public void testAlgorithmWithNoApproximation1() throws RoutePlannerException {
        FactoryMap factoryMap = FactoryMapUtil.createDemoMap();
        RoutePlanner planner = new RoutePlanner(factoryMap, factoryMap.getPoint(0, 0),
                factoryMap.getPoint(factoryMap.getSizeX() - 1, factoryMap.getSizeY() - 1), false);
        RouteUtil.computeShortestPath(planner, null, 1);
        assertEquals("With No Approximation all fields should be explored", 0,
                countNumOfState(planner.getFactoryMap(), FactoryMap.MapPoint.State.UNEXPLORED));
    }
}
